﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.GeneralizedQuadraticAssignment.Common;

namespace HeuristicLab.Problems.GeneralizedQuadraticAssignment {
  [Item("SlackMinimizationSolutionCreator", "A heuristic that creates a solution to the Generalized Quadratic Assignment Problem by minimizing the amount of slack.")]
  [StorableClass]
  public class SlackMinimizationSolutionCreator : GQAPStochasticSolutionCreator {

    public IValueLookupParameter<IntValue> MaximumTriesParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["MaximumTries"]; }
    }
    public IValueLookupParameter<BoolValue> CreateMostFeasibleSolutionParameter {
      get { return (IValueLookupParameter<BoolValue>)Parameters["CreateMostFeasibleSolution"]; }
    }
    public IValueLookupParameter<IntValue> DepthParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["Depth"]; }
    }

    [StorableConstructor]
    protected SlackMinimizationSolutionCreator(bool deserializing) : base(deserializing) { }
    protected SlackMinimizationSolutionCreator(SlackMinimizationSolutionCreator original, Cloner cloner) : base(original, cloner) { }
    public SlackMinimizationSolutionCreator()
      : base() {
      Parameters.Add(new ValueLookupParameter<IntValue>("MaximumTries", "The maximum number of tries to create a feasible solution after which an exception is thrown. If it is set to 0 or a negative value there will be an infinite number of attempts to create a feasible solution.", new IntValue(100000)));
      Parameters.Add(new ValueLookupParameter<BoolValue>("CreateMostFeasibleSolution", "If this is set to true the operator will always succeed, and outputs the solution with the least violation instead of throwing an exception.", new BoolValue(false)));
      Parameters.Add(new ValueLookupParameter<IntValue>("Depth", "How deep the algorithm should look forward.", new IntValue(3)));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new SlackMinimizationSolutionCreator(this, cloner);
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      if (!Parameters.ContainsKey("Depth")) {
        Parameters.Add(new ValueLookupParameter<IntValue>("Depth", "How deep the algorithm should look forward.", new IntValue(3)));
      }
    }

    public static IntegerVector CreateSolution(IRandom random, DoubleArray demands, DoubleArray capacities, int depth, int maximumTries, bool createMostFeasibleSolution, CancellationToken cancel) {
      IntegerVector result = null;
      bool isFeasible = false;
      int counter = 0;
      double minViolation = double.MaxValue;
      var slack = new DoubleArray(capacities.Length);
      var assignment = new Dictionary<int, int>(demands.Length);

      while (!isFeasible) {
        cancel.ThrowIfCancellationRequested();
        if (maximumTries > 0) {
          counter++;
          if (counter > maximumTries) {
            if (createMostFeasibleSolution) break;
            else throw new InvalidOperationException("A feasible solution could not be obtained after " + maximumTries + " attempts.");
          }
        }
        assignment.Clear();
        for (int i = 0; i < capacities.Length; i++) slack[i] = capacities[i];
        var remainingEquipment = new HashSet<int>(Enumerable.Range(0, demands.Length));
        while (remainingEquipment.Any()) {
          var equipment = remainingEquipment.ChooseProportionalRandom(random, 1, x => demands[x], true, true).Single();
          remainingEquipment.Remove(equipment);

          var freeLocations = GetLocationsByFillLevel(equipment, demands[equipment], slack);
          if (!freeLocations.Any()) break;
          double bestSlack = double.MaxValue;
          int bestLocation = -1;
          int[] additionalEquipment = new int[0];
          foreach (var location in freeLocations.Shuffle(random)) {
            HashSet<int> newEquipment;
            var estimatedSlack = EstimatePotentialSlack(location.Key, location.Value, remainingEquipment, demands, out newEquipment, depth);
            if (estimatedSlack < bestSlack || (estimatedSlack == bestSlack && newEquipment.Count < additionalEquipment.Length)) {
              bestSlack = estimatedSlack;
              bestLocation = location.Key;
              additionalEquipment = newEquipment.ToArray();
            }
          }
          assignment[equipment] = bestLocation;
          slack[bestLocation] -= demands[equipment];
          foreach (var eq in additionalEquipment) {
            remainingEquipment.Remove(eq);
            assignment[eq] = bestLocation;
            slack[bestLocation] -= demands[eq];
          }
        }
        if (assignment.Count != demands.Length) {
          // complete the solution
          while (remainingEquipment.Any()) {
            var f = remainingEquipment.ChooseMax(x => demands[x]);
            var l = Enumerable.Range(0, capacities.Length).ChooseMax(x => slack[x]);
            remainingEquipment.Remove(f);
            assignment.Add(f, l);
            slack[l] -= demands[f];
          }
        }
        double violation = GQAPEvaluator.EvaluateOverbooking(slack, capacities);
        isFeasible = violation == 0;
        if (isFeasible || violation < minViolation) {
          result = new IntegerVector(assignment.OrderBy(x => x.Key).Select(x => x.Value).ToArray());
          minViolation = violation;
        }
      }
      return result;
    }

    private static double EstimatePotentialSlack(int location, double slack, HashSet<int> remainingEquipment, DoubleArray demands, out HashSet<int> addedEquipment, int depth = 3) {
      var feasibleEquipment = remainingEquipment.Where(x => demands[x] < slack).ToArray();
      addedEquipment = new HashSet<int>();

      if (!feasibleEquipment.Any()) return slack;
      if (depth == 0) {
        var e = feasibleEquipment.ChooseMax(x => demands[x]);
        addedEquipment.Add(e);
        return slack - demands[e];
      }

      double bestSlack = slack;
      int minSlackEquipment = -1;
      foreach (var e in feasibleEquipment) {
        remainingEquipment.Remove(e);
        HashSet<int> childEquipment;
        var potentialSlack = EstimatePotentialSlack(location, slack - demands[e], remainingEquipment, demands, out childEquipment, depth - 1);
        if (bestSlack > potentialSlack || (bestSlack == potentialSlack && childEquipment.Count < addedEquipment.Count)) {
          bestSlack = potentialSlack;
          minSlackEquipment = e;
          addedEquipment.Clear();
          foreach (var child in childEquipment) addedEquipment.Add(child);
        }
        remainingEquipment.Add(e);
      }
      if (minSlackEquipment > -1)
        addedEquipment.Add(minSlackEquipment);
      return bestSlack;
    }

    protected override IntegerVector CreateRandomSolution(IRandom random, DoubleArray demands, DoubleArray capacities) {
      return CreateSolution(random, demands, capacities,
        DepthParameter.ActualValue.Value,
        MaximumTriesParameter.ActualValue.Value,
        CreateMostFeasibleSolutionParameter.ActualValue.Value,
        CancellationToken);
    }

    private static IEnumerable<KeyValuePair<int, double>> GetLocationsByFillLevel(int equipment, double demand, DoubleArray slack) {
      var freeLocations = slack
        .Select((v, idx) => new KeyValuePair<int, double>(idx, v))
        .Where(x => x.Value >= demand);
      if (!freeLocations.Any()) {
        return Enumerable.Empty<KeyValuePair<int, double>>();
      }
      return freeLocations.Select(x => new KeyValuePair<int, double>(x.Key, slack[x.Key] - demand)).OrderBy(x => x.Value);
    }
  }
}
